from preprocspectra.transformations import snv, sg, msc, plus_sg, area_norm

import pandas as pd


def make_transformations(datasets, transformations, index_spectra_start):
    if not isinstance(datasets, list):
        raise ValueError('datasets should be a list of dataframes.')

    if not isinstance(transformations, list):
        raise ValueError('transformations should be a list of strings.')

    for dataset in datasets:
        if not isinstance(dataset, pd.DataFrame):
            raise ValueError('dataset should be a pandas dataframe.')

    ALL = ['RAW', 'SNV', 'MSC', 'AREA_NORM', 'SG_11', 'SG_25', 'SG2_11', 'SG2_25', 
           'SNV_SG11', 'SNV_SG2_11', 'SNV_SG25', 'SNV_SG2_25', 'SG11_SNV',
           'SG2_11_SNV', 'SG2_25_SNV', 'SG11_MSC', 'SG2_11_MSC', 'SG25_MSC',
           'SG2_25_MSC', 'AREA_NORM_SG2_11', 'AREA_NORM_SG2_25', 'MSC_SG2_11', 
           'MSC_SG2_25']
    
    transformations = [x.upper() for x in transformations]
    dataset_transformed = []

    if 'ALL' in transformations:
        transformations = ALL
    
    if 'RAW' in transformations:
        if len(datasets) == 2:
            dataset_transformed.append((datasets[0], datasets[1], 'RAW'))
        if len(datasets) == 1:
            dataset_transformed.append((datasets[0], 'RAW'))

    if 'SNV' in transformations:
        if len(datasets) == 2:
            df_snv = snv(datasets[0], spectra_start=index_spectra_start)
            df_val_snv = snv(datasets[1], spectra_start=index_spectra_start)
            dataset_transformed.append((df_snv, df_val_snv, 'SNV'))

        elif len(datasets) == 1:
            df_snv = snv(datasets[0], spectra_start=index_spectra_start)
            dataset_transformed.append((df_snv, 'SNV'))
    
    
    
    if 'MSC' in transformations:
        if len(datasets) == 2:
            df_msc = msc(datasets[0], spectra_start=index_spectra_start)
            df_val_msc = msc(datasets[1], spectra_start=index_spectra_start)
            dataset_transformed.append((df_msc, df_val_msc, 'MSC'))
        
        elif len(datasets) == 1:
            df_msc = msc(datasets[0], spectra_start=index_spectra_start)
            dataset_transformed.append((df_msc, 'MSC'))
    


    if 'AREA_NORM' in transformations:
        if len(datasets) == 2:
            df_areanorm = area_norm(datasets[0], spectra_start=index_spectra_start)
            df_val_areanorm = area_norm(datasets[1], spectra_start=index_spectra_start)
            dataset_transformed.append((df_areanorm, df_val_areanorm, 'AREA_NORM'))
        elif len(datasets) == 1:
            df_areanorm = area_norm(datasets[0], spectra_start=index_spectra_start)
            dataset_transformed.append((df_areanorm, 'AREA_NORM'))
    


    if 'SG_11' in transformations:
        if len(datasets) == 2:
            df_sg11 = sg(datasets[0], differentiation=1, window_size=11, spectra_start=index_spectra_start)
            df_val_sg11 = sg(datasets[1], differentiation=1, window_size=11, spectra_start=index_spectra_start)
            dataset_transformed.append((df_sg11, df_val_sg11, 'SG_11'))
        elif len(datasets) == 1:
            df_sg11 = sg(datasets[0], differentiation=1, window_size=11, spectra_start=index_spectra_start)
            dataset_transformed.append((df_sg11, 'SG_11'))
    

    
    if 'SG_25' in transformations:
        if len(datasets) == 2:
            df_sg25 = sg(datasets[0], differentiation=1, window_size=25, spectra_start=index_spectra_start)
            df_val_sg25 = sg(datasets[1], differentiation=1, window_size=25, spectra_start=index_spectra_start)
            dataset_transformed.append((df_sg25, df_val_sg25, 'SG_25'))
        elif len(datasets) == 1:
            df_sg25 = sg(datasets[0], differentiation=1, window_size=25, spectra_start=index_spectra_start)
            dataset_transformed.append((df_sg25, 'SG_25'))
    


    if 'SG2_11' in transformations:
        if len(datasets) == 2:
            df_sg2_11 = sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start)
            df_val_sg2_11 = sg(datasets[1], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start)
            dataset_transformed.append((df_sg2_11, df_val_sg2_11, 'SG2_11'))
        if len(datasets) == 1:
            df_sg2_11 = sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start)
            dataset_transformed.append((df_sg2_11, 'SG2_11'))
    


    if 'SG2_25' in transformations:
        if len(datasets) == 2:
            df_sg2_25 = sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start)
            df_val_sg2_25 = sg(datasets[1], differentiation=2, window_size=25, spectra_start=index_spectra_start)
            dataset_transformed.append((df_sg2_25, df_val_sg2_25, 'SG2_25'))
        if len(datasets) == 1:
            df_sg2_25 = sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start)
            dataset_transformed.append((df_sg2_25, 'SG2_25'))
    


    if 'SNV_SG11' in transformations:
        if len(datasets) == 2:
            df_snv_sg11 = plus_sg(datasets[0], differentiation=1, window_size=11, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            df_val_snv_sg11 = plus_sg(datasets[1], differentiation=1, window_size=11, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            dataset_transformed.append((df_snv_sg11, df_val_snv_sg11, 'SNV_SG11'))
        if len(datasets) == 1:
            df_snv_sg11 = plus_sg(datasets[0], differentiation=1, window_size=11, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            dataset_transformed.append((df_snv_sg11, 'SNV_SG11'))
    


    if 'SNV_SG2_11' in transformations:
        if len(datasets) == 2:
            df_snv_sg2_11 = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            df_val_snv_sg2_11 = plus_sg(datasets[1], differentiation=2, window_size=11, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            dataset_transformed.append((df_snv_sg2_11, df_val_snv_sg2_11, 'SNV_SG2_11'))
        if len(datasets) == 1:
            df_snv_sg2_11 = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            dataset_transformed.append((df_snv_sg2_11, 'SNV_SG2_11'))
    


    if 'SNV_SG25' in transformations:
        if len(datasets) == 2:
            df_snv_sg25 = plus_sg(datasets[0], differentiation=1, window_size=25, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            df_val_snv_sg25 = plus_sg(datasets[1], differentiation=1, window_size=25, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            dataset_transformed.append((df_snv_sg25, df_val_snv_sg25, 'SNV_SG25'))
        if len(datasets) == 1:
            df_snv_sg25 = plus_sg(datasets[0], differentiation=1, window_size=25, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            dataset_transformed.append((df_snv_sg25, 'SNV_SG25'))
    


    if 'SNV_SG2_25' in transformations:
        if len(datasets) == 2:
            df_snv_sg2_25 = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            df_val_snv_sg2_25 = plus_sg(datasets[1], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            dataset_transformed.append((df_snv_sg2_25, df_val_snv_sg2_25, 'SNV_SG2_25'))
        if len(datasets) == 1:
            df_snv_sg2_25 = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=snv, sg_first=False)
            dataset_transformed.append((df_snv_sg2_25, 'SNV_SG2_25'))
    


    if 'SG11_SNV' in transformations:
        if len(datasets) == 2:        
            df_sg11_snv = plus_sg(datasets[0], differentiation=1, window_size=11, spectra_start=index_spectra_start, transformation=snv, sg_first=True)
            df_val_sg11_snv = plus_sg(datasets[1], differentiation=1, window_size=11, spectra_start=index_spectra_start, transformation=snv, sg_first=True)
            dataset_transformed.append((df_sg11_snv, df_val_sg11_snv, 'SG11_SNV'))
        if len(datasets) == 1:
            df_sg11_snv = plus_sg(datasets[0], differentiation=1, window_size=11, spectra_start=index_spectra_start, transformation=snv, sg_first=True)
            dataset_transformed.append((df_sg11_snv, 'SG11_SNV'))
    


    if 'SG2_11_SNV' in transformations:
        if len(datasets) == 2:        
            df_sg2_11_snv = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=snv, sg_first=True)
            df_val_sg2_11_snv = plus_sg(datasets[1], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=snv, sg_first=True)
            dataset_transformed.append((df_sg2_11_snv, df_val_sg2_11_snv, 'SG2_11_SNV'))
        if len(datasets) == 1:
            df_sg2_11_snv = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=snv, sg_first=True)
            dataset_transformed.append((df_sg2_11_snv, 'SG2_11_SNV'))
    


    if 'SG2_25_SNV' in transformations:
        if len(datasets) == 2:
            df_sg2_25_snv = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=snv, sg_first=True)
            df_val_sg2_25_snv = plus_sg(datasets[1], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=snv, sg_first=True)
            dataset_transformed.append((df_sg2_25_snv, df_val_sg2_25_snv, 'SG2_25_SNV'))
        if len(datasets) == 1:
            df_sg2_25_snv = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=snv, sg_first=True)
            dataset_transformed.append((df_sg2_25_snv, 'SG2_25_SNV'))
    


    if 'SG11_MSC' in transformations:
        if len(datasets) == 2:
            df_sg11_msc = plus_sg(datasets[0], differentiation=1, window_size=11, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            df_val_sg11_msc = plus_sg(datasets[1], differentiation=1, window_size=11, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            dataset_transformed.append((df_sg11_msc, df_val_sg11_msc, 'SG11_MSC'))
        if len(datasets) == 1:
            df_sg11_msc = plus_sg(datasets[0], differentiation=1, window_size=11, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            dataset_transformed.append((df_sg11_msc, 'SG11_MSC'))




    if 'SG2_11_MSC' in transformations:
        if len(datasets) == 2:
            df_sg2_11_msc = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            df_val_sg2_11_msc = plus_sg(datasets[1], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            dataset_transformed.append((df_sg2_11_msc, df_val_sg2_11_msc, 'SG2_11_MSC'))
        if len(datasets) == 1:
            df_sg2_11_msc = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            dataset_transformed.append((df_sg2_11_msc, 'SG2_11_MSC'))




    if 'SG25_MSC' in transformations:
        if len(datasets) == 2:
            df_sg25_msc = plus_sg(datasets[0], differentiation=1, window_size=25, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            df_val_sg25_msc = plus_sg(datasets[1], differentiation=1, window_size=25, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            dataset_transformed.append((df_sg25_msc, df_val_sg25_msc, 'SG25_MSC'))
        if len(datasets) == 1:
            df_sg25_msc = plus_sg(datasets[0], differentiation=1, window_size=25, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            dataset_transformed.append((df_sg25_msc, 'SG25_MSC'))
    


    if 'SG2_25_MSC' in transformations:
        if len(datasets) == 2:
            df_sg2_25_msc = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            df_val_sg2_25_msc = plus_sg(datasets[1], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            dataset_transformed.append((df_sg2_25_msc, df_val_sg2_25_msc, 'SG2_25_MSC'))
        if len(datasets) == 1:
            df_val_sg2_25_msc = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=msc, sg_first=True)
            dataset_transformed.append((df_val_sg2_25_msc, 'SG2_25_MSC'))
    


    if 'AREA_NORM_SG2_11' in transformations:
        if len(datasets) == 2:
            df_normalize_sg2_11 = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=area_norm, sg_first=False)
            df_val_normalize_sg2_11 = plus_sg(datasets[1], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=area_norm, sg_first=False)
            dataset_transformed.append((df_normalize_sg2_11, df_val_normalize_sg2_11, 'AREA_NORM_SG2_11'))
        if len(datasets) == 1:
            df_normalize_sg2_11 = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=area_norm, sg_first=False)
            dataset_transformed.append((df_normalize_sg2_11, 'AREA_NORM_SG2_11'))
    


    if 'AREA_NORM_SG2_25' in transformations:
        if len(datasets) == 2:
            df_normalize_sg2_25 = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=area_norm, sg_first=False)
            df_val_normalize_sg2_25 = plus_sg(datasets[1], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=area_norm, sg_first=False)
            dataset_transformed.append((df_normalize_sg2_25, df_val_normalize_sg2_25, 'AREA_NORM_SG2_25'))
        if len(datasets) == 1:
            df_normalize_sg2_25 = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=area_norm, sg_first=False)
            dataset_transformed.append((df_normalize_sg2_25, 'AREA_NORM_SG2_25'))
    


    if 'MSC_SG2_11' in transformations:
        if len(datasets) == 2:
            df_msc_sg2_11 = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=msc, sg_first=False)
            df_val_msc_sg2_11 = plus_sg(datasets[1], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=msc, sg_first=False)
            dataset_transformed.append((df_msc_sg2_11, df_val_msc_sg2_11, 'MSC_SG2_11'))
        if len(datasets) == 1:
            df_msc_sg2_11 = plus_sg(datasets[0], differentiation=2, window_size=11, polynominal_order=2, spectra_start=index_spectra_start, transformation=msc, sg_first=False)
            dataset_transformed.append((df_msc_sg2_11, 'MSC_SG2_11'))

    

    if 'MSC_SG2_25' in transformations:
        if len(datasets) == 2:
            df_msc_sg2_25 = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=msc, sg_first=False)
            df_val_msc_sg2_25 = plus_sg(datasets[1], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=msc, sg_first=False)
            dataset_transformed.append((df_msc_sg2_25, df_val_msc_sg2_25, 'MSC_SG2_25'))
        if len(datasets) == 1:
            df_msc_sg2_25 = plus_sg(datasets[0], differentiation=2, window_size=25, spectra_start=index_spectra_start, transformation=msc, sg_first=False)
            dataset_transformed.append((df_msc_sg2_25, 'MSC_SG2_25'))
    

    return dataset_transformed



import pandas as pd
import numpy as np
from scipy.signal import savgol_filter

def snv(dataset, spectra_start):
    
    """
    Apply the standard normal variate transformation.
        - the dataset should be a dataframe
        - spectra_start is the column that spectra start
    """

    if not isinstance(dataset, pd.DataFrame):
        raise ValueError('dataset should be a pandas dataframe')
    if type(spectra_start) not in [int]:
        raise ValueError('spectra_start should be a integer that reference a column')

    df = dataset.copy()

    avg = df.iloc[: ,spectra_start:].mean(axis=1)
    stds = df.iloc[: ,spectra_start:].std(axis=1)

    df.iloc[:, spectra_start:] = df.iloc[:, spectra_start:].subtract(avg, axis=0)
    df.iloc[:, spectra_start:] = df.iloc[:, spectra_start:].divide(stds, axis=0)
    
    return df


def msc(dataset, spectra_start):

    """
    Apply the Multiplicative Scatter Correction transformation.
        - the dataset should be a dataframe
        - spectra_start is the column that spectra start
    """

    if not isinstance(dataset, pd.DataFrame):
        raise ValueError('dataset should be a pandas dataframe')
    if type(spectra_start) not in [int]:
        raise ValueError('spectra_start should be a integer that reference a column')

    df = dataset.copy()

    spectra = df.iloc[:, spectra_start:]

    ref = np.mean(df.iloc[:, spectra_start:], axis=0)
    mean = df.iloc[:, spectra_start:].mean(axis=1)

    df.iloc[:, spectra_start:] = df.iloc[:, spectra_start:].subtract(mean, axis=0)

    for i in range(spectra.shape[0]):
        
        fit = np.polyfit(ref, df.iloc[i, spectra_start :].astype('float64'), deg=1, full=True)
        
        df.iloc[i, spectra_start:] = ((df.iloc[i, spectra_start :] - fit[0][1]) / fit[0][0]).values

    return df


def area_norm(dataset, spectra_start):
    
    """
    Apply the Area Normalize.
        - the dataset should be a dataframe
        - spectra_start is the column that spectra start
    """

    if not isinstance(dataset, pd.DataFrame):
        raise ValueError('dataset should be a pandas dataframe')
    if type(spectra_start) not in [int]:
        raise ValueError('spectra_start should be a integer that reference a column')

    df = dataset.copy()

    sum_of_rows = df.iloc[:, spectra_start:].sum(axis=1)

    df.iloc[:, spectra_start:] = df.iloc[:, spectra_start:].divide(sum_of_rows, axis=0)
    
    return df



def sg(dataset, differentiation, window_size, spectra_start, polynominal_order=4):
    
    """
    Apply the Savitzky-Golay filter.
        - the dataset should be a dataframe
        - differentiation is the derivative order
        - window_size is a window size (must be odd).
        - polynominal_order for equation
    """
    if not isinstance(dataset, pd.DataFrame):
        raise ValueError('dataset should be a pandas dataframe')
    if type(differentiation) not in [int]:
        raise ValueError('differentiation should be a integer')
    if type(window_size) not in [int]:
        raise ValueError('window_size should be a integer')
    if type(polynominal_order) not in [int]:
        raise ValueError('polynominal_order should be a integer')
    if type(spectra_start) not in [int]:
        raise ValueError('spectra_start should be a integer that reference a column')

    df = dataset.copy()

    sg_df = savgol_filter(df.iloc[:, spectra_start:], window_length=window_size, polyorder=polynominal_order, deriv=differentiation, axis=-1)

    sg_df = pd.DataFrame(sg_df)

    sg_df.columns = df.iloc[:, spectra_start:].columns
    sg_df.index = df.iloc[:, spectra_start:].index

    df.iloc[:, spectra_start:] = sg_df.iloc[:, spectra_start:]

    gap = window_size // 2
    
    columns_sg = list(sg_df.iloc[:,gap:-gap].columns)
    columns_dataset = list(df.iloc[:,spectra_start:].columns)

    columns_for_drop = list(set(columns_dataset) - set(columns_sg))

    df = df.drop(columns_for_drop, axis=1)

    return df

def plus_sg(dataset, sg_first, transformation, differentiation, window_size, spectra_start, polynominal_order=4):
    
    """
    Apply the Savitzky-Golay filter after apply SNV transformation.
        - the dataset should be a dataframe
        - sg_first should be a boolean value. True if sg should  be performed before the additional transformation
        - differentiation is the derivative order
        - window_size is a window size (must be odd).
        - polynominal_order for equation
    """

    if sg_first not in [True, False]:
        raise ValueError('sg_first should be a boolean value')
    if not isinstance(dataset, pd.DataFrame):
        raise ValueError('dataset should be a pandas dataframe')
    if type(differentiation) not in [int]:
        raise ValueError('differentiation should be a integer')
    if type(window_size) not in [int]:
        raise ValueError('window_size should be a integer')
    if type(polynominal_order) not in [int]:
        raise ValueError('polynominal_order should be a integer')
    if type(spectra_start) not in [int]:
        raise ValueError('spectra_start should be a integer that reference a column')
    
    df = dataset.copy()
    if sg_first == True:
        df_first = sg(df, differentiation, window_size, spectra_start, polynominal_order)
        df_transformed = transformation(df_first, spectra_start=spectra_start)
    else:
        df_first = transformation(df, spectra_start=spectra_start)
        df_transformed = sg(df_first, differentiation, window_size, spectra_start, polynominal_order)

    return df_transformed



__all__ = ['snv', 'sg', 'plus_sg', 'area_norm', 'msc', 'make_transformations']

from preprocspectra.transformations import snv
from preprocspectra.transformations import sg
from preprocspectra.transformations import plus_sg
from preprocspectra.transformations import area_norm
from preprocspectra.transformations import msc
from preprocspectra.handle_transformations import make_transformations 



import setuptools

setuptools.setup(
	name="preprocspectra",
	version="0.0.1",
    author="Diego José da Silva",
    author_email="diegojsilvabr@gmail.com",
    description="A simple library to apply preprocessing of spectra obtained from infrared spectroscopy.",
    url="https://github.com/dijsilva/spectra-transform",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'pandas>=1.0.3',
        'numpy>=1.18.3',
        'scipy>=1.4.1',
    ],
    python_requires='>=3.6'
)
